import scala.annotation.tailrec

@tailrec
def foldLeft[A, B](list: List[A])(init: B)(f: (B, A) => B): B = {
  list match {
    case Nil => init
    case head::tail => foldLeft(tail)(f(init, head))(f)
  }
}

@tailrec
def dropWhile[T](list: List[T])(predicate: T => Boolean): List[T] = {
  list match {
    case head::tail if predicate(head) => dropWhile(tail)(predicate)
    case _ => list
  }
}

def predicates[T](list: List[T])(preds: List[T => Boolean]): List[T] = {
  list.foldRight(List.empty[T])((e: T, l: List[T]) => {
    //if (preds.foldLeft(true)((a, b) => a && b(e))) e::l
    if (preds.forall(b => b(e))) e::l
    else l
  })
}

def fixedPoint(f: Int => Int): Int => Int = {
  @tailrec
  def func(x: Int): Int = {
    val y: Int = f(x)
    if (x == y) y
    else func(y)
  }
  func
}

fixedPoint(x => x / 2 + 5)(40)